package ru.dbotthepony.mc.otm.worldgen.placement

import com.mojang.serialization.MapCodec
import com.mojang.serialization.codecs.RecordCodecBuilder
import net.minecraft.core.BlockPos
import net.minecraft.util.RandomSource
import net.minecraft.util.valueproviders.FloatProvider
import net.minecraft.util.valueproviders.IntProvider
import net.minecraft.world.level.levelgen.placement.PlacementContext
import net.minecraft.world.level.levelgen.placement.PlacementModifier
import net.minecraft.world.level.levelgen.placement.PlacementModifierType
import ru.dbotthepony.mc.otm.core.math.component1
import ru.dbotthepony.mc.otm.core.math.component2
import ru.dbotthepony.mc.otm.core.math.component3
import ru.dbotthepony.mc.otm.core.math.minus
import ru.dbotthepony.mc.otm.core.math.plus
import ru.dbotthepony.mc.otm.registry.data.MPlacementModifiers
import java.util.stream.Stream
import kotlin.math.PI
import kotlin.math.roundToInt

/**
 * Regular ellipsoid ("cloud") placement, suitable to be used as non-primary placement modifier
 *
 * This placement modifier is designed to be terminal; other placement modifiers after this MUST NOT be placed
 */
data class EllipsoidPlacement(
	/**
	 * X position within ellipsoid sampler, expected to be in range -1 .. 1
	 */
	val x: FloatProvider,

	/**
	 * Z position within ellipsoid sampler, expected to be in range -1 .. 1
	 */
	val z: FloatProvider,

	/**
	 * Y position within ellipsoid sampler, expected to be in range -1 .. 1
	 */
	val y: FloatProvider,

	/**
	 * Total amount of block positions to generate
	 */
	val count: IntProvider,

	/**
	 * Ellipsoid size sampler on X axis
	 */
	val xLength: FloatProvider,

	/**
	 * Ellipsoid size sampler on Z axis
	 */
	val zLength: FloatProvider,

	/**
	 * Ellipsoid size sampler on Y axis
	 */
	val yLength: FloatProvider,
) : PlacementModifier() {
	init {
		require(xLength.minValue >= 1f) { "Bad ellipsoid x minimal size: $xLength" }
		require(zLength.minValue >= 1f) { "Bad ellipsoid z minimal size: $zLength" }
		require(yLength.minValue >= 1f) { "Bad ellipsoid y minimal size: $yLength" }
	}

	override fun getPositions(
		context: PlacementContext,
		random: RandomSource,
		position: BlockPos
	): Stream<BlockPos> {
		val count = count.sample(random)

		if (count <= 0)
			return Stream.empty()

		val xLength = xLength.sample(random)
		val zLength = zLength.sample(random)
		val yLength = yLength.sample(random)

		val xPow = xLength * xLength
		val zPow = zLength * zLength
		val yPow = yLength * yLength

		return Stream.generate {
			val x = this.x.sample(random) * xLength
			val y = this.y.sample(random) * yLength
			val z = this.z.sample(random) * zLength
			BlockPos(x.toInt(), y.toInt(), z.toInt())
		}
			.limit(count.toLong() * 10)
			.filter {
				val (ellipsoidX, ellipsoidY, ellipsoidZ) = it

				(ellipsoidX * ellipsoidX) / xPow +
					(ellipsoidY * ellipsoidY) / yPow +
					(ellipsoidZ * ellipsoidZ) / zPow <= 1.0f
			}
			.distinct()
			.limit(count.toLong())
			.map { it + position }
	}

	override fun type(): PlacementModifierType<*> {
		return MPlacementModifiers.ELLIPSOID_PLACEMENT
	}

	companion object {
		val CODEC: MapCodec<EllipsoidPlacement> by lazy {
			RecordCodecBuilder.mapCodec {
				it.group(
					FloatProvider.CODEC.fieldOf("x").forGetter(EllipsoidPlacement::x),
					FloatProvider.CODEC.fieldOf("y").forGetter(EllipsoidPlacement::y),
					FloatProvider.CODEC.fieldOf("z").forGetter(EllipsoidPlacement::z),
					IntProvider.CODEC.fieldOf("count").forGetter(EllipsoidPlacement::count),
					FloatProvider.codec(1f, Float.MAX_VALUE).fieldOf("x_size").forGetter(EllipsoidPlacement::xLength),
					FloatProvider.codec(1f, Float.MAX_VALUE).fieldOf("z_size").forGetter(EllipsoidPlacement::zLength),
					FloatProvider.codec(1f, Float.MAX_VALUE).fieldOf("y_size").forGetter(EllipsoidPlacement::yLength),
				).apply(it, ::EllipsoidPlacement)
			}
		}
	}
}
